###### Functions for Exp3


##########
# the Exp 3 algorithm
# n= size of the sample, l, vector of losses, eta the learning parameter
##########

Exp3 <- function(n,l,eta){
  
  K=length(l) # the number of arms is the size of the losses

  # initialize the list of arms, it is of length n.
  
  Listofarmschosen=rep(0,n) 
  
  #initialize the matrix of probabilities
  mat_prob = matrix(0,ncol = K, nrow = n)
  #initialize the probability vector
  p=rep(1/K,K)
  mat_prob[1,] = p
  
  #initialize the list of cumulative losses
  cumulativelosses = rep(0,K)
  
  # first arm picked at random
  I =sample(1:K, 1, prob=p, replace=TRUE)
  Listofarmschosen[1]=I
  
  for (t in 2:n)
  {
    cumulativelosses[I] = cumulativelosses[I] + l[I]/p[I]
    
    W = sum(exp(-eta*cumulativelosses))
    p = exp(-eta*cumulativelosses)/W
    
    mat_prob[t,] = p
    

    #sample an action with respect to the distribution p
    I = sample(1:K, 1, prob=p, replace=TRUE)
    
    #update the list of arms chosen
    Listofarmschosen[t]=I
  }
  list(Listofarmschosen,mat_prob)
}


##########
# Given a list of arms, prob_delta returns the list of the updated probabilities for a parameter delta,
# as well as the time until one the updated probability for an arm reaches epsilon.
# n= size of the sample, l, vector of losses, eta the learning parameter.
##########



prob_delta <- function(arms,delta,l,epsilon=0){
  n=length(arms)
  K=length(l)
  #initialize the list of cumulative losses
  cumulativelosses = rep(0, K)
  
  #initialize the matrix of probabilities
  pdelta = matrix(0,nrow = n,ncol = K)
  pdelta[1,] = rep(1/K, K)

  for (t in 2:n)
  {
    I = arms[t-1]
    
    cumulativelosses[I] =cumulativelosses[I] + l[I]/ pdelta[t-1,I]
    
    #compute the sum of the weights
    W = sum(exp(-delta*cumulativelosses))
    
    #compute the new probability distribution over arms
    pdelta[t,] = exp(-delta*cumulativelosses)/W
  }
  
  leps=length(epsilon)
  Tmax=rep(0,leps)
  for(e in 1:leps)
  {  
    ind=which(pdelta<epsilon[e],arr.ind=TRUE)
    if(length(ind)==0)
    {
       Tmax[e]=n+1
    }else{
      Tmax[e]=min(ind[,1])
    }
  }  
  list(pdelta,Tmax)
}

##########
# The function loglikelihood computes the loglikelihood for a parameter delta.
# l, vector of losses, arms, the list of arms generated by the Exp3 procedure for a parameter eta.
##########

loglikelihood<- function(l,delta,arms)
{
  #initialize the probability distribution pdelta
  
  pdelta = prob_delta(arms,delta,l)
  prob=pdelta[[1]]
  
  n=length(arms)
  
  #initialize the log likelihood 
  ll = 0
  for (t in 1:n){
    I = arms[t]
    proba = prob[t,I]
    #update the log likelihood
    ll = ll + log(proba)
  }
  return(ll)
}


##########
# The function MLE computes the maximum likelihood estimator for a list of parameters range.
# l, vector of losses, arms, the list of arms generated by the Exp3 procedure for a parameter eta.
##########


MLE<-function(arms,l,range)
{
  fn <-function(delta){
    -loglikelihood(l,delta,arms)
  }
  opti=DEoptim(fn,range[1],range[2],DEoptim.control(itermax = 50))
  estimator =opti$optim$bestmem[[1]] 
  return(estimator)
}  


##########
# The function MLE computes the maximum likelihood estimator stopped at epsilon for a list of parameters range.
# l, vector of losses, arms, the list of arms generated by the Exp3 procedure for a parameter eta.
##########

MLE_truncated<-function(arms,l,range,epsilon,nbreaks)
{
  grid_delta=seq(range[1],range[2],length.out=nbreaks)
  ld=length(grid_delta)
  Tmax=rep(0,ld)
  for(id in 1:ld)
  {  
    Tmax[id]=prob_delta(arms,grid_delta[id],l,epsilon=epsilon)[[2]]
  }
  Tmax_min=min(Tmax)
  
  fn <-function(delta){
    -loglikelihood(l,delta,arms[1:(Tmax_min-1)])
  }
  opti=DEoptim(fn,range[1],range[2],DEoptim.control(itermax = 50))
  estimator =opti$optim$bestmem[[1]] 
  return(list(estimator,Tmax_min))
} 